/*	===========================================================================
 UPC - EETAC - CSD - EMC - http://digsys.upc.edu

 * LAB6: Analogue inputs, analogue sensors, design step #2
 * Documentation location: https://digsys.upc.edu/csd/EED/lab06/lab06.html
 * Hardware for prototyping: Arduino Uno board. Chip ATmega328P
 * Similar circuit: (step #1)	

Planning: From step #1, let us calculate mean values


=============================================================================*/


/*	===========================================================================
	Definitions
	======================================================================== */
// Switches 
const byte pin_VM 	= 4;  // Mean value measurement switch is SW0

const byte pin_AD0	= A0;  // Analogue inputs
const byte pin_AD1  = A1;
const byte pin_AD2  = A2;

// Start-stop push-button. Pin PD3 is also the microcontroller's external INT1
const byte INT1_pin_ST_BUTTON = 3;  // 

// LEDs for indicating the current state 
const byte pin_LED2 = 9; 
const byte pin_LED1 = 6; 
const byte pin_LED0 = 5; 

// LED for generating the 1 Hz squared waveform when acquiring
const byte pin_SLED = 17; 

 
// FSM list of states
// FSM list of states 
const char IDLE           = 'A'; 
const char SET_ACQ_TIMER  = 'B';
const char Do_NOTHING     = 'C';
const char ACQUIRE        = 'D';
const char TEMP_MEAS      = 'E';
const char TRANSMIT       = 'F';
const char STOP_ACQ_TIMER = 'G';


// Default timer period to sample analogue inputs 
const unsigned long SAMPLING_PERIOD = 499; // 500 ms   (2 Hz sampling frequency)

/*	===========================================================================
	Global RAM variables
	======================================================================== */

byte   	var_VM; 		// switch to be read
					
byte   	var_S; 			// For representing the state in binary
byte   	var_SLED; 		// Squared waveform when running

char 	var_current_state ;    // This is the key variable: FSM state memory

byte 	var_SP_flag; 	// Sample period flag 
byte 	var_ST_flag;    // It will be set when one click detected on interrupt

byte 	var_USART_flag; // For writing or sending ASCII messages only when 
						// new information

//The system will poll internal timer when acquiring
byte   	var_TMR_flag;


static unsigned int i;  //measurement string index to save current measururements


// Voltage variables are recorded in a circular string
static unsigned int var_Vs0[10];   
static unsigned int var_Vs1[10]; 
static unsigned int var_Vs2[10];   

// voltage mean values
static unsigned int var_Vs0_m;
static unsigned int var_Vs1_m;
static unsigned int var_Vs2_m; 

// Temperature variables can be instantaneous or calculated as a mean value
static float var_Ts0; 
static float var_Ts1;   
static float var_Ts2;  


// For timing the square waveform: 
static unsigned long var_previous_time;   
static unsigned long var_current_time;    // current time in ms
static unsigned long var_time_elapsed;   //  time elapsed since last flag

static unsigned int var_sampling_period; 




/*	===========================================================================
	System initialisation
	=========================================================================*/
void setup() { 
// Configuring digital output pins for indicating the current state in binary
// for debugging purposes
	pinMode(pin_LED2, OUTPUT);  
	pinMode(pin_LED1, OUTPUT);	
	pinMode(pin_LED0, OUTPUT);	

// Output where to connect the sampling LED that will toggle every Ts	
	pinMode(pin_SLED, OUTPUT);	

// Configuring digital input pin for reading VM 
	pinMode(pin_VM, INPUT);  

// This is the special pin to be an external interrupt source INT1:
	pinMode(INT1_pin_ST_BUTTON, INPUT_PULLUP); 
	
// Serial monitor initialisation
	Serial.begin(9600); 
 
//Initialise the FSM
	var_current_state = IDLE;

// Keep the timer off until it is required when acquiring temperature
	var_TMR_flag = 0;
	var_current_time = 0; 
	var_previous_time = 0; 

		// Serial monitor ASCII message:
	Serial.println("Project LAB6 - EMC");
	Serial.println("Temperature acquisition"); 
	Serial.println("System IDLE: waiting for ST click to start"); 
	var_USART_flag = 0;

//	---------------------------------------------------------------------------
// This is the interrupt handler. Read the Arduino documentation at: 
// https://www.arduino.cc/reference/en/language/functions/external-interrupts/attachinterrupt/
// attachInterrupt(digitalPinToInterrupt(pin), ISR, mode) (recommended)
	
	attachInterrupt(digitalPinToInterrupt(INT1_pin_ST_BUTTON), interrupt_service_routine_INT1, FALLING);
	// Interrupt when detecting a falling edge to start and stop operations
}

/*	===========================================================================
	Infinite loop 
	======================================================================== */
void loop() {       
	read_inputs(); 		// capture the switch voltage values into RAM variables
	state_logic();    // calculating the next state to go
	output_logic();		// calculating outputs 
	write_outputs(); // Convert RAM variable values into pin output voltage
}

/*	===========================================================================
	State logic to determine where to go next
	======================================================================== */
void state_logic(void){

	update_time_flag();    //  to keep track of real time and be able to 
							// acquire signals at sampling frequency

	switch (var_current_state) {
	case IDLE:	// Do nothing, wait for start signal
		if (var_ST_flag == 1) {	  // start the process
			  var_current_state = SET_ACQ_TIMER;
			  var_ST_flag = 0;	  //Clear flag when one used
		}
		  else
			var_current_state = IDLE;     
		break;

	case SET_ACQ_TIMER:	// Do nothing, wait for start signal
		var_current_state = Do_NOTHING;       
		break;

	case Do_NOTHING:	// Wait for sampling period
		if (var_SP_flag == 1) {	  // start acquiring 
			var_current_state = ACQUIRE;
			var_SP_flag = 0;	  //Clear flag when one used
		}
		  else
			var_current_state = Do_NOTHING;      
		break;

	case ACQUIRE:
		var_current_state = TEMP_MEAS;     
		break;

	case TEMP_MEAS:	
		var_current_state = TRANSMIT;  
		break;

	case TRANSMIT:
		if (var_ST_flag == 1) {	  // stop acquiring 
			var_current_state = STOP_ACQ_TIMER;
			var_ST_flag = 0;	  //Clear flag when one used
		}
		  else
			var_current_state = Do_NOTHING;      
		break;
		
	case STOP_ACQ_TIMER:	// go to Idle to wait for another Start pulse
		var_current_state = IDLE;     
		break;

	}	
}

/*	===========================================================================
	Output logic 
	======================================================================== */
void output_logic(void){
int var_buff, j;
	
	switch (var_current_state) {
	case IDLE:
		var_SLED = 0;
		var_S = 0;		

		// Serial monitor ASCII message:
		if (var_USART_flag == 1){
			Serial.println("System IDLE: waiting for ST click to start"); 
			var_USART_flag = 0;
		}

		break;

	case SET_ACQ_TIMER:	
		var_SLED = 0;
		var_S = 1;	
		var_sampling_period = SAMPLING_PERIOD; 		
		for (i = 0; i <10; i++){  // init voltage buffer
			var_Vs0[i] = 0;	var_Vs1[i] = 0;	var_Vs2[i] = 0;
		}
		i = 0; 
		// Turn on the timer for counting sampling periods
		var_TMR_flag = 1;
		var_current_time = 0; 
		var_previous_time = 0; 	 
		break;

	case Do_NOTHING: // waiting for sampling period ...
		var_S = 2;	

		break;	


	case ACQUIRE: 
		// Toggle operation with a XOR simply to indicate running acquisition
		var_SLED = var_SLED^0b00000001; 
		var_S = 3;	
		
			// Three consecutive analogue to digital conversions
		var_Vs0[i] = analogRead(pin_AD0); // read V_AD0 voltage
		var_Vs1[i] = analogRead(pin_AD1); // read V_AD1 voltage
		var_Vs2[i] = analogRead(pin_AD2); // read V_AD2 voltage

		if (var_VM == 0){   // using instantaneous values
			var_Vs0_m = var_Vs0[i];
			var_Vs1_m = var_Vs1[i];
			var_Vs2_m = var_Vs2[i];
		}
		else{  // using mean values
			var_Vs0_m = 0; var_Vs1_m = 0; var_Vs2_m = 0;
			for (j = 0; j <10; j++){
				var_Vs0_m = var_Vs0[j] + var_Vs0_m;
				var_Vs1_m = var_Vs1[j] + var_Vs1_m;
				var_Vs2_m = var_Vs2[j] + var_Vs2_m;
			}
			var_Vs0_m = var_Vs0_m/10;
			var_Vs1_m = var_Vs1_m/10;
			var_Vs2_m = var_Vs2_m/10;

		}

		i++ ; // circular buffer
		if (i == 10) i = 0;

		break;	


	case TEMP_MEAS:	// measurement equation
		var_S = 4;	

// Temperatures
		var_Ts0 = (var_Vs0_m - 307.2)/10.24; 
		var_Ts1 = (var_Vs1_m - 307.2)/10.24; 	
		var_Ts2 = (var_Vs2_m - 307.2)/10.24; 

	// Rounding a float with only one decimal digit
		var_Ts0 = var_Ts0*10 + 0.5;
		var_buff = (int)var_Ts0;
		var_Ts0 = (float)var_buff/10;

		var_Ts1 = var_Ts1*10 + 0.5;
		var_buff = (int)var_Ts1;
		var_Ts1 = (float)var_buff/10;

		var_Ts2 = var_Ts2*10 + 0.5;
		var_buff = (int)var_Ts2;
		var_Ts2 = (float)var_buff/10;

		break;
			
	case TRANSMIT:	// measurement equation	
		var_S = 5;	
	
		// Serial monitor:
		Serial.print("Ts0 = "); 	
		Serial.print((float)var_Ts0, 1); 		
		Serial.print(" degC     "); 

		Serial.print("Ts1 = "); 	
		Serial.print((float)var_Ts1, 1); 		
		Serial.print(" degC     "); 

		Serial.print("Ts2 = "); 	
		Serial.print((float)var_Ts2, 1); 		
		Serial.println(" degC"); 

		break;

	case STOP_ACQ_TIMER:
		var_S = 6;	

// Turn off the timer for counting sampling periods
		var_TMR_flag = 0;
		var_current_time = 0; 
		var_previous_time = 0; 	 

// For transmitting one time the IDLE message
		var_USART_flag = 1;	
		break;
	
	}
	
}	


/*	===========================================================================
	ISR external interrupt INT0 pin 
This function is executed when an active edge is detected at pin PD2 (INT0)
The idea is simply set the global variable
	======================================================================== */
void interrupt_service_routine_INT1 (void){

	var_ST_flag = 1;  

	// LAB experimentation on debouncing techniques: 
	//delay(5); // wait some time 

}


/*	===========================================================================
	read_inputs 
	======================================================================== */
void read_inputs(void){

   if (digitalRead(pin_VM) == 1) // Active-low connected switch
		var_VM = 1;
   else
		var_VM = 0;
   
}

/*	===========================================================================
	write_outputs 
	======================================================================== */
void write_outputs(void){
byte var_buf;  // local variable for saving partial results

	var_buf = var_S & 0b00000001; //Masking the bit of interest
   if (var_buf == 1)
       digitalWrite(pin_LED0, LOW); // Active-low connected LED 
   else
       digitalWrite(pin_LED0, HIGH);  
	
	var_buf = (var_S & 0b00000010)>>1;
   if (var_buf == 1)
       digitalWrite(pin_LED1, LOW); // Active-low connected LED 
   else
       digitalWrite(pin_LED1, HIGH);  

	var_buf = (var_S & 0b00000100)>>2;
   if (var_buf == 1)
       digitalWrite(pin_LED2, LOW); // Active-low connected LED 
   else
       digitalWrite(pin_LED2, HIGH);  
	

   if (var_SLED == 1)
       digitalWrite(pin_SLED, LOW); // Active-low connected LED 
   else
       digitalWrite(pin_SLED, HIGH);  

}



/*	===========================================================================
	Establishing the sampling frequency using internal timer
	======================================================================== */
void update_time_flag(void){

// Count real time only when acquiring
	if (var_TMR_flag == 1){
	  // From internal CLK: 
		var_current_time = millis();
		var_time_elapsed =  var_current_time - var_previous_time;  

		if(var_time_elapsed  > var_sampling_period) {
				// is the flag is set is because time elapsed is sampling period
			var_SP_flag = 1;   
				// to start another sampling period cycle
			var_previous_time = var_current_time; 
		}
	}
}



